# -*- coding: utf-8 -*-

'''Layout depends on Blocks and Shapes.

**Layout** here refers to the content and position of text, image and table. The target is to convert
source blocks and shapes to a *flow layout* that can be re-created as docx elements like paragraph and
table. In this library, The page structure/layout is maintained by ``TableBlock``. So, detecting and 
parsing table block is the principle steps.

The prerequite work is done before this step:

1. Clean source blocks and shapes in Page level. The main step is to merge blocks 
   horizontally considering flow layout (only one block in horizontal direction).
#. Parse structure in document level, e.g. page header/footer.
#. Parse Section and Column layout in Page level. 

The page layout parsing idea:

1. Parse table layout in Column level.
    (a) Detect explicit tables first based on shapes. 
    (#) Then, detect stream tables based on original text blocks and parsed explicit tables.
    (#) Move table contained blocks (text block or explicit table) to associated cell-layout.
#. Parse text format for text blocks in current layout.
#. Repeat above steps for cell-layout in parsed table level.
'''

from ..common import constants
from ..text.TextBlock import TextBlock
from ..shape.Shapes import Shapes


class Layout:
    '''Blocks and shapes structure and formats.'''

    def __init__(self, blocks=None, shapes=None):
        ''' Initialize layout.

        Args:
            blocks (Blocks): Blocks representing text/table contents.
            shapes (Shapes): Shapes representing table border, shading and text style like underline, highlight.
            parent (Page, Column, Cell): The object that this layout belonging to.
        '''
        from .Blocks import Blocks # avoid import conflicts
        from ..table.TablesConstructor import TablesConstructor

        self.blocks = Blocks(instances=blocks, parent=self)
        self.shapes = Shapes(instances=shapes, parent=self)        
        self._table_parser = TablesConstructor(parent=self) # table parser


    def working_bbox(self, *args, **kwargs):
        '''Working bbox of current Layout.'''
        raise NotImplementedError


    def contains(self, *args, **kwargs):
        '''Whether given element is contained in this layout.'''
        raise NotImplementedError


    def store(self):
        '''Store parsed layout in dict format.'''
        return {
            'blocks': self.blocks.store(),
            'shapes': self.shapes.store()
        }


    def restore(self, data:dict):
        '''Restore Layout from parsed results.'''
        self.blocks.restore(data.get('blocks', []))
        self.shapes.restore(data.get('shapes', []))
        return self


    def assign_blocks(self, blocks:list):
        '''Add blocks to this layout. 
        
        Args:
            blocks (list): a list of text/table block to add.
        
        .. note::
            If a text block is partly contained, it must deep into line -> span -> char.
        '''
        for block in blocks: self._assign_block(block)


    def assign_shapes(self, shapes:list):
        '''Add shapes to this cell. 
        
        Args:
            shapes (list): a list of Shape instance to add.
        '''
        # add shape if contained in cell
        for shape in shapes:
            if self.working_bbox & shape.bbox: self.shapes.append(shape)


    def parse(self, **settings):
        '''Parse layout.

        Args:
            settings (dict): Layout parsing parameters.
        '''
        # parse tables
        self._parse_table_layout(**settings)

        # improve layout after table parsing
        self._improve_layout(**settings)

        # parse text format in current layout
        self._parse_text_format(**settings)

        # parse sub-layout, i.e. cell layouts under table block
        for block in filter(lambda e: e.is_table_block, self.blocks):
            block.parse(**settings)


    def _assign_block(self, block):
        '''Add block to this layout. 
        
        Args:
            block (TextBlock, TableBlock): Text/table block to add. 
        '''
        # add block directly if fully contained in cell
        if self.contains(block, threshold=constants.FACTOR_SAME):
            self.blocks.append(block)
            return
        
        # add nothing if no intersection
        if not self.bbox & block.bbox: return

        # otherwise, further check lines in text block
        if not block.is_text_image_block:  return
        
        # NOTE: add each line as a single text block to avoid overlap between table block and combined lines
        split_block = TextBlock()
        lines = [line.intersects(self.bbox) for line in block.lines]
        split_block.add(lines)
        self.blocks.append(split_block)


    def _parse_table_layout(self, **settings):
        '''Parse table layout: 
        
        * detect explicit tables first based on shapes, 
        * then stream tables based on original text blocks and parsed explicit tables;
        * move table contained blocks (text block or explicit table) to associated cell layout.
        '''
        # merge blocks horizontally, e.g. remove overlap blocks.
        # NOTE: It's to merge blocks in physically horizontal direction, i.e. without considering text direction.
        self.blocks.join_horizontally(False, 
                        settings['line_overlap_threshold'],
                        settings['line_merging_threshold'])
        
        # parse table structure/format recognized from explicit shapes
        if settings['parse_lattice_table']:
            self._table_parser.lattice_tables(
                            settings['connected_border_tolerance'],
                            settings['min_border_clearance'],
                            settings['max_border_width'])
        
        # parse table structure based on implicit layout of text blocks
        if settings['parse_stream_table']:
            self._table_parser.stream_tables(
                            settings['min_border_clearance'],
                            settings['max_border_width'],
                            settings['float_layout_tolerance'],
                            settings['line_separate_threshold'])
    

    def _improve_layout(self, **settings):
        '''Adjust layout after table parsing:

        * split blocks in current level back to original layout if possible
        * merge adjacent and similar blocks in vertical direction
        '''
        # blocks are joined horizontally in table parsing stage, now change back to original layout
        self.blocks.split_back(
            settings['float_layout_tolerance'], 
            settings['line_separate_threshold'])
        
        # one paragraph may be split in multiple blocks by `PyMuPDF`, now merge them together
        # by checking vertical distance
        self.blocks.join_vertically_by_space(
            settings['block_merging_threshold']
        )

        # one block may consist of multiple real paragraphs by `PyMuPDF`, now split them back
        # by checking text
        self.blocks.split_vertically_by_text(
            settings['line_break_free_space_ratio'],
            settings['new_paragraph_free_space_ratio']
        )
    

    def _parse_text_format(self, **settings):
        '''Parse text format, e.g. text highlight, paragraph indentation. 
        '''
        # parse text format, e.g. highlight, underline
        self.blocks.parse_text_format(self.shapes.text_style_shapes)
        
        # paragraph / line spacing
        self.blocks.parse_spacing(
                        settings['line_separate_threshold'],
                        settings['line_break_width_ratio'],
                        settings['line_break_free_space_ratio'],
                        settings['lines_left_aligned_threshold'],
                        settings['lines_right_aligned_threshold'],
                        settings['lines_center_aligned_threshold'],
                        settings['line_condense_spacing'])


# -*- coding: utf-8 -*-

'''A group of Line objects.
'''

import logging
import string
from .Line import Line
from .TextSpan import TextSpan
from ..image.ImageSpan import ImageSpan
from ..common.Collection import ElementCollection
from ..common.share import TextAlignment
from ..common import constants


class Lines(ElementCollection):
    '''Collection of text lines.'''

    @property
    def unique_parent(self):
        '''Whether all contained lines have same parant.'''
        if not bool(self): return False

        first_line = self._instances[0]
        return all(line.same_source_parent(first_line) for line in self._instances)


    def append(self, line:Line):
        """Override. Append a line and update line pid and parent bbox.

        Args:
            line (Line): Target line to add.
        """
        super().append(line)

        # update original parent id
        if not self._parent is None:
            line.pid = id(self._parent)


    def restore(self, raws:list):
        '''Construct lines from raw dicts list.'''
        for raw in raws:
            line = Line(raw)
            self.append(line)
        return self


    @property
    def image_spans(self):
        '''Get all ImageSpan instances.'''
        spans = []
        for line in self._instances:
            spans.extend(line.image_spans)
        return spans

    
    def join(self, line_overlap_threshold:float, line_merging_threshold:float):
        '''Merge lines aligned horizontally, e.g. make inline image as a span in text line.'''
        # skip if empty
        if not self._instances: return self

        # valid to merge lines on condition that every tow lines are in same row
        def valid_joining_lines(line, candidates):
            return all(line.in_same_row(_line) for _line in candidates)
        
        # merge lines
        def get_merged_line(candidates):
            line = candidates[0]
            for c_line in candidates[1:]:
                line.add(c_line.spans)
            return line

        # sort lines
        self.sort()

        # check each line
        lines = Lines()
        candidates = [self._instances[0]] # first line
        for i in range(1, len(self._instances)):
            pre_line, line = self._instances[i-1], self._instances[i]
           
            # ignore this line if overlap with previous line
            if line.get_main_bbox(pre_line, threshold=line_overlap_threshold):
                logging.warning('Ignore Line "%s" due to overlap', line.text)
                continue

            # add line directly if not aligned horizontally with previous line
            if not line.in_same_row(pre_line):
                to_join_line = False

            # if it exists x-distance obviously to previous line,
            # take it as a separate line as it is
            elif abs(line.bbox.x0-pre_line.bbox.x1) > line_merging_threshold:
                to_join_line = False 

            # now, this line will be append to previous line as a span
            else:
                to_join_line = True

            # add line directly
            if not to_join_line:
                # merge candidate lines (if any)
                if candidates: lines.append(get_merged_line(candidates))
                candidates = []

                # add this line
                lines.append(line)
            
            # prepare for merging lines: valid
            elif valid_joining_lines(line, candidates):
                candidates.append(line)
            
            # prepare for merging lines: invalid -> add each line directly
            else:
                # release candidate lines
                for c_line in candidates: lines.append(c_line)
                candidates = []

                # add this line
                lines.append(line)

        # NOTE: in case last group
        if candidates: lines.append(get_merged_line(candidates))

        # update lines in block
        self.reset(lines)


    def split_back(self):
        '''Split lines into groups, in which all lines are from same original text block.

        Returns:
            list: A list of Lines contained in same original text block.
        '''
        fun = lambda a,b: a.same_source_parent(b)
        groups = self.group(fun)

        # NOTE: group() may destroy the order of lines, so sort in line level
        for group in groups: group.sort()

        return groups

    
    def split_vertically_by_text(self, line_break_free_space_ratio:float, new_paragraph_free_space_ratio:float):
        '''Split lines into separate paragraph, because ``PyMuPDF`` stores lines in ``block``,
        rather than real paragraph.

        .. note::
            Considered only normal reading direction, from left to right, from top
            to bottom.
        '''
        rows = self.group_by_physical_rows()

        # skip if only one row
        num = len(rows)
        if num==1: return rows

        # standard row width with first row excluded, considering potential indentation of fist line
        W = max(row[-1].bbox[2]-row[0].bbox[0] for row in rows[1:])
        H = sum(row[0].bbox[3]-row[0].bbox[1] for row in rows) / num

        # check row by row
        res = []
        lines = Lines()
        punc = tuple(constants.SENTENSE_END_PUNC)
        start_of_para = end_of_para = False # start/end of paragraph
        start_of_sen = end_of_sen = False   # start/end of sentense
        for row in rows:
            end_of_sen = row[-1].text.strip().endswith(punc)
            w =  row[-1].bbox[2]-row[0].bbox[0]

            # end of a sentense and free space at the end -> end of paragraph
            if end_of_sen and w/W <= 1.0-line_break_free_space_ratio:
                end_of_para = True

            # start of sentense and free space at the start -> start of paragraph
            elif start_of_sen and (W-w)/H >= new_paragraph_free_space_ratio:
                start_of_para = True

            # take action
            if end_of_para:
                lines.extend(row)
                res.append(lines)
                lines = Lines()
            elif start_of_para:
                res.append(lines)
                lines = Lines()
                lines.extend(row)
            else:
                lines.extend(row)

            # for next round
            start_of_sen = end_of_sen
            start_of_para = end_of_para = False
        
        # close the action
        if lines: res.append(lines)

        return res


    def strip(self, delete_end_line_hyphen:bool):
        '''Remove redundant blanks of each line and update bbox accordingly.'''
        # strip each line and update bbox: 
        # keep at least one blank at both sides in case extra blanks existed
        strip_status = []
        strip_status.extend([line.strip() for line in self._instances])
        stripped = any(strip_status)
        if stripped: self._parent.update_bbox(self.bbox) # update bbox        

        # word process:
        # - it might miss blank between words from adjacent lines
        # - it's optional to delete hyphen since it might not at the line end
        #   after conversion

        punc_ex_hyphen = ''.join(c for c in string.punctuation if c!='-')
        def is_end_of_english_word(c):
            return c.isalnum() or (c and c in punc_ex_hyphen)
        
        for i, line in enumerate(self._instances[:-1]):
            # last char in this line
            end_span = line.spans[-1]
            if not isinstance(end_span, TextSpan): continue
            end_chars = end_span.chars
            if not end_chars: continue 
            end_char = end_chars[-1]

            # first char in next line
            start_span = self._instances[i+1].spans[0]
            if not isinstance(start_span, TextSpan): continue
            start_chars = start_span.chars
            if not start_chars: continue 
            next_start_char = start_chars[0]            

            # delete hyphen if next line starts with lower case letter
            if delete_end_line_hyphen and \
                end_char.c.endswith('-') and next_start_char.c.islower(): 
                end_char.c = '' # delete hyphen in a tricky way


            # add a space if both the last char and the first char in next line are alphabet,  
            # number, or English punctuation (excepting hyphen)
            if is_end_of_english_word(end_char.c) and is_end_of_english_word(next_start_char.c):
                end_char.c += ' ' # add blank in a tricky way
            
        return stripped


    def sort(self):
        '''Sort lines considering text direction.

        Taking natural reading direction for example: reading order for rows, from left to 
        right for lines in row.

        In the following example, A should come before B.

        ::

                         +-----------+
            +---------+  |           |
            |   A     |  |     B     |
            +---------+  +-----------+
        
        Steps:

            * Sort lines in reading order, i.e. from top to bottom, from left to right.
            * Group lines in row.
            * Sort lines in row: from left to right.
        '''
        # sort in reading order
        self.sort_in_reading_order()

        # split lines in separate row
        lines_in_rows = [] # type: list[list[Line]]

        for line in self._instances:

            # add lines to a row group if not in same row with previous line
            if not lines_in_rows or not line.in_same_row(lines_in_rows[-1][-1]):
                lines_in_rows.append([line])
            
            # otherwise, append current row group
            else:
                lines_in_rows[-1].append(line)
        
        # sort lines in each row: consider text direction
        idx = 0 if self.is_horizontal_text else 3
        self._instances = []
        for row in lines_in_rows:
            row.sort(key=lambda line: line.bbox[idx])
            self._instances.extend(row)


    def is_flow_layout(self, float_layout_tolerance:float, line_separate_threshold:float):
        '''Check if flow layout. 
        
        A flow layout satisfy condition that lines in each physical row have:
        
        * same original text block
        * enough overlap in vertical direction.
        * no significant gap between adjacent two lines.
        '''
        # group lines in same row
        fun = lambda a, b: a.horizontally_align_with(b, factor=float_layout_tolerance) and \
                            not a.vertically_align_with(b, factor=constants.FACTOR_ALMOST) 
        groups = self.group(fun)        
        
        # check each row
        idx0, idx1 = (0, 2) if self.is_horizontal_text else (3, 1)
        for lines in groups:
            num = len(lines)
            if num==1: continue

            # same original parent
            if not all(line.same_source_parent(lines[0]) for line in lines):
                return False

            # check vertical overlap
            if not all(line.in_same_row(lines[0]) for line in lines):
                return False

            # check distance between lines
            for i in range(1, num):
                dis = abs(lines[i].bbox[idx0]-lines[i-1].bbox[idx1])
                if dis >= line_separate_threshold: return False

        return True


    def parse_text_format(self, shape):
        '''Parse text format with style represented by rectangle shape.
        
        Args:
            shape (Shape): Potential style shape applied on blocks.
        
        Returns:
            bool: Whether a valid text style.
        '''
        flag = False

        for line in self._instances:
            # any intersection in this line?
            intsec = shape.bbox & line.get_expand_bbox(constants.MAJOR_DIST)
            
            if not intsec: 
                if shape.bbox.y1 < line.bbox.y0: break # lines must be sorted in advance
                continue

            # yes, then try to split the spans in this line
            split_spans = []
            for span in line.spans: 
                # include image span directly
                if isinstance(span, ImageSpan): split_spans.append(span)                   

                # split text span with the format rectangle: span-intersection-span
                else:
                    spans = span.split(shape, line.is_horizontal_text)
                    split_spans.extend(spans)
                    flag = True
                                            
            # update line spans                
            line.spans.reset(split_spans)

        return flag


    def parse_line_break(self, bbox, 
                line_break_width_ratio:float, 
                line_break_free_space_ratio:float,
                line_condense_spacing:float):
        '''Whether hard break each line. In addition, condense charaters at end of line to avoid unexpected 
        line break. PDF sets precisely width of each word, here just an approximation to set condense spacing
        for last two words.

        Args:
            bbox (Rect): bbox of parent layout, e.g. page or cell.
            line_break_width_ratio (float): user defined threshold, break line if smaller than this value.
            line_break_free_space_ratio (float): user defined threshold, break line if exceeds this value.
            line_condense_spacing (float): user defined total condensed spacing to avoid unexpected line break.

        Hard line break helps ensure paragraph structure, but pdf-based layout calculation may
        change in docx due to different rendering mechanism like font, spacing. For instance, when
        one paragraph row can't accommodate a Line, the hard break leads to an unnecessary empty row.
        Since we can't 100% ensure a same structure, it's better to focus on the content - add line
        break only when it's necessary to, e.g. short lines.
        '''

        block = self.parent        
        idx0, idx1 = (0, 2) if block.is_horizontal_text else (3, 1)
        block_width = abs(block.bbox[idx1]-block.bbox[idx0])
        layout_width = bbox[idx1] - bbox[idx0]

        # hard break if exceed the width ratio
        line_break = block_width/layout_width <= line_break_width_ratio

        # check by each physical row
        rows = self.group_by_physical_rows()
        single_row = len(rows)==1
        for lines in rows:
            # ----------------------------
            # line break
            # ----------------------------
            for line in lines: line.line_break = 0

            # check the end line depending on text alignment
            if block.alignment == TextAlignment.RIGHT:
                end_line = lines[0]
                free_space = abs(block.bbox[idx0]-end_line.bbox[idx0])
            else:
                end_line = lines[-1]
                free_space = abs(block.bbox[idx1]-end_line.bbox[idx1])
            
            if block.alignment == TextAlignment.CENTER: free_space *= 2 # two side space
            
            # break line if 
            # - width ratio lower than the threshold; or 
            # - free space exceeds the threshold
            if line_break or free_space/block_width > line_break_free_space_ratio:
                end_line.line_break = 1

            # ----------------------------
            # character spacing
            # ----------------------------
            row_width = abs(lines[-1].bbox[idx1]-lines[0].bbox[idx0])
            if block_width-row_width>constants.MINOR_DIST: continue
            last_span = lines[-1].spans[-1]
            if isinstance(last_span, TextSpan) and not single_row: 
                # condense characters if negative value
                last_span.condense_spacing = line_condense_spacing

        
        # no break for last row
        for line in rows[-1]: line.line_break = 0


    def parse_tab_stop(self, line_separate_threshold:float):
        '''Calculate tab stops for parent block and whether add TAB stop before each line. 

        Args:
            line_separate_threshold (float): Don't need a tab stop if the line gap less than this value.
        '''
        # set all tab stop positions for parent block
        # Note these values are relative to the left boundary of parent block
        block = self.parent        
        idx0, idx1 = (0, 2) if block.is_horizontal_text else (3, 1)
        fun = lambda line: round(abs(line.bbox[idx0]-block.bbox[idx0]), 1)
        all_pos = set(map(fun, self._instances))
        block.tab_stops = list(filter(lambda pos: pos>=constants.MINOR_DIST, all_pos))

        # no tab stop need
        if not block.tab_stops: return

        # otherwise, set tab stop option for each line
        ref = block.bbox[idx0]
        for i, line in enumerate(self._instances):
            # left indentation implemented with tab
            distance = line.bbox[idx0] - ref
            if distance>line_separate_threshold:
                line.tab_stop = 1

            # update stop reference position
            if line==self._instances[-1]: break
            ref = line.bbox[idx1] if line.in_same_row(self._instances[i+1]) else block.bbox[idx0]


# -*- coding: utf-8 -*-

'''
Objects representing PDF path (stroke and filling) extracted by ``page.getDrawings()``.

This method is new since ``PyMuPDF`` 1.18.0, with both pdf raw path and annotations like Line, 
Square and Highlight considered.

* https://pymupdf.readthedocs.io/en/latest/page.html#Page.getDrawings
* https://pymupdf.readthedocs.io/en/latest/faq.html#extracting-drawings
'''

import fitz
from ..common.share import lazyproperty
from ..common.Collection import  Collection
from .Path import Path


class Paths(Collection):
    '''A collection of paths.'''

    def restore(self, raws:list):
        '''Initialize paths from raw data get by ``page.getDrawings()``.'''
        rect = (0, 0, self.parent.width, self.parent.height)
        for raw in raws:
            path = Path(raw)
            # ignore path out of page
            if not path.bbox.intersects(rect): continue
            self.append(path)
        
        return self
    
    @lazyproperty
    def bbox(self):
        bbox = fitz.Rect()
        for instance in self._instances: bbox |= instance.bbox
        return bbox


    def contains_curve(self, ratio:float):
        """Whether any curve paths exist with threshold considered. 

        **Criterion**: the ratio of non-iso-oriented paths to total count >= ``ratio``.

        Args:
            ratio (float): Threshold for non-iso-oriented paths.

        Returns:
            bool: Whether any curve paths exist.
        """
        if not self._instances: return False
        curve_paths = list(filter(lambda path: not path.is_iso_oriented, self._instances))
        return len(curve_paths)/len(self._instances) > ratio


    def plot(self, page):
        """Plot paths for debug purpose.

        Args:
            page (fitz.Page): ``PyMuPDF`` page.
        """
        if not self._instances: return
        # make a drawing canvas and plot path
        canvas = page.newShape()
        for path in self._instances: path.plot(canvas)
        canvas.commit() # commit the drawing shapes to page
    

    def _to_iso_paths(self):
        """Convert contained paths to ISO strokes and rectangular fills.

        Returns:
            list: A list of ``Shape`` raw dicts.

        .. note::
            Non-ISO path is ignored.
        """
        paths = []
        for path in self._instances:
            # consider iso-oriented path only
            if not path.is_iso_oriented: continue
            paths.extend(path.to_shapes())
        return paths


    def to_shapes(self, curve_path_ratio:float=0.2):
        """Extract ISO-oriented (horizontal/vertical) paths for table parsing.

        Args:
            curve_path_ratio (float, optional): Not iso-oriented paths if exceed this ratio. Defaults to 0.2.

        Returns:
            tuple: (Shape raw dicts, shape bbox list, whether exists vector graphics).
        """
        # group connected paths -> each group is a potential vector graphic        
        paths_group = self.group_by_connectivity(dx=0.1, dy=0.1)

        iso_paths, iso_areas, exist_svg = [], [], False
        for paths in paths_group:
            # keep potential table border paths
            if paths.contains_curve(curve_path_ratio):
                exist_svg = True
            else:
                iso_paths.extend(paths._to_iso_paths())
                iso_areas.extend([path.bbox for path in paths])

        return iso_paths, iso_areas, exist_svg




# -*- coding: utf-8 -*-

'''Extract images from PDF.

Both raster images and vector graphics are considered:

* Normal images like jpeg or png could be extracted with method ``page.getText('rawdict')`` 
  and ``Page.getImageList()``. Note the process for png images with alpha channel.
* Vector graphics are actually composed of a group of paths, represented by operators like
  ``re``, ``m``, ``l`` and ``c``. They're detected by finding the contours with ``opencv``.
'''

import fitz
from ..common.share import BlockType
from ..common.Collection import Collection
from ..common.Element import Element


class ImagesExtractor:
    def __init__(self, page:fitz.Page) -> None:
        '''Extract images from PDF page.
        
        Args:
            page (fitz.Page): pdf page to extract images.
        '''
        self._page = page


    def extract_images(self, clip_image_res_ratio:float=3.0):
        '''Extract normal images with ``Page.getImageList()``.

        Args:
            clip_image_res_ratio (float, optional): Resolution ratio of clipped bitmap. Defaults to 3.0.

        Returns:
            list: A list of extracted and recovered image raw dict.
        
        .. note::
            ``Page.getImageList()`` contains each image only once, which may less than the real count of images in a page.
        '''
        # pdf document
        doc = self._page.parent

        # check each image item:
        # (xref, smask, width, height, bpc, colorspace, ...)
        images = []
        for item in self._page.getImageList(full=True):
            # should always wrap getImageBbox in a try-except clause, per
            # https://github.com/pymupdf/PyMuPDF/issues/487
            try:
                item = list(item)
                item[-1] = 0
                bbox = self._page.getImageBbox(item) # item[7]: name entry of such an item
            except ValueError:
                continue

            # ignore images outside page
            if not bbox.intersects(self._page.rect): continue

            # recover image
            pix = self._recover_pixmap(doc, item)

            # regarding images consist of alpha values only, i.e. colorspace is None,
            # the turquoise color shown in the PDF is not part of the image, but part of PDF background.
            # So, just to clip page pixmap according to the right bbox
            # https://github.com/pymupdf/PyMuPDF/issues/677
            if not pix.colorspace:
                pix = self._clip_page(bbox, zoom=clip_image_res_ratio)

            raw_dict = self._to_raw_dict(pix, bbox)
            images.append(raw_dict)
        return images
    

    def extract_vector_graphics(self, exclude_areas:list, clip_image_res_ratio:float=3.0):
        '''Detect and extract vector graphics by clipping associated page area.

        Args:
            exclude_areas (list): A list of bbox-like ``(x0, y0, x1, y1)`` area to exclude, 
                e.g. raster image area, table area.
            clip_image_res_ratio (float, optional): Resolution ratio of clipped bitmap. Defaults to 3.0.

        Returns:
            list: A list of extracted and recovered image raw dict.
        
        .. note::
            Contours for vector graphics are detected first with ``opencv-python``.
        '''
        # find contours
        contours = self._detect_svg_contours(exclude_areas)

        # filter contours
        fun = lambda a,b: a.bbox & b.bbox
        groups = contours.group(fun)

        # clip images
        images = []
        for i,group in enumerate(groups):
            bbox = group.bbox
            pix = self._page.getPixmap(clip=bbox, matrix=fitz.Matrix(clip_image_res_ratio, clip_image_res_ratio))
            # pix.save(f'{i}.png')
            raw_dict = self._to_raw_dict(pix, bbox)
            images.append(raw_dict)
        
        return images


    def _detect_svg_contours(self, exclude_areas:list):
        '''Find contour of potential vector graphics.'''
        import cv2 as cv
        import numpy as np

        # clip page and convert to opencv image
        img_byte = self._clip_page(zoom=1.0).getPNGData()
        src = cv.imdecode(np.frombuffer(img_byte, np.uint8), cv.IMREAD_COLOR)        

        # gray and exclude areas      
        gray = cv.cvtColor(src, cv.COLOR_BGR2GRAY)
        for bbox in exclude_areas:
            x0, y0, x1, y1 = map(int, bbox)
            gray[y0-2:y1+2, x0-2:x1+2] = 255
            # cv.rectangle(src, (x0,y0), (x1,y1), (0,255,0), 1) # plot excluded areas

        # Gauss blur and binary        
        gauss = cv.GaussianBlur(gray, (9,9), 0)
        _, binary = cv.threshold(gauss, 250, 255, cv.THRESH_BINARY_INV)        
        
        # detect contour with opencv
        rows, cols = binary.shape[0:2]
        kernel = cv.getStructuringElement(cv.MORPH_RECT, (cols//50, rows//50))
        dst = cv.morphologyEx(binary, cv.MORPH_CLOSE, kernel) # close
        contours, hierarchy = cv.findContours(dst,cv.RETR_EXTERNAL,cv.CHAIN_APPROX_SIMPLE)
        
        collection = Collection()
        for contour in contours:  
            x, y, w, h = cv.boundingRect(contour)
            e = Element().update_bbox((x,y,x+w,y+h))
            collection.append(e)
            # cv.rectangle(src, (x,y), (x+w,y+h), (255,0,0), 1)  # detected svg bbox
        
        # plot detected images for debug
        # cv.imshow("img", src)
        # cv.waitKey(0)

        return collection


    @staticmethod
    def _to_raw_dict(image:fitz.Pixmap, bbox:fitz.Rect):
        '''Store Pixmap ``image`` to raw dict.

        Args:
            image (fitz.Pixmap): Pixmap to store.
            bbox (fitz.Rect): Boundary box the pixmap.

        Returns:
            dict: Raw dict of the pixmap.
        '''
        return {
            'type': BlockType.IMAGE.value,
            'bbox': tuple(bbox),
            'ext': 'png',
            'width': image.width,
            'height': image.height,
            'image': image.getPNGData()
        }


    def _hide_page_text(self):
        '''Hide page text before clipping page.'''
        # render Tr: set the text rendering mode
        # - 3: neither fill nor stroke the text -> invisible
        # read more:
        # - https://github.com/pymupdf/PyMuPDF/issues/257
        # - https://www.adobe.com/content/dam/acom/en/devnet/pdf/pdfs/pdf_reference_archives/PDFReference.pdf
        doc = self._page.parent # type: fitz.Document

        # NOTE: contents referenced in this page
        for (xref, name, invoker, bbox) in self._page.get_xobjects():
            stream = doc.xref_stream(xref).replace(b'BT', b'BT 3 Tr') \
                                             .replace(b'Tm', b'Tm 3 Tr') \
                                             .replace(b'Td', b'Td 3 Tr')
            doc.update_stream(xref, stream)

        # direct page content
        for xref in self._page.get_contents():
            stream = doc.xref_stream(xref).replace(b'BT', b'BT 3 Tr') \
                                             .replace(b'Tm', b'Tm 3 Tr') \
                                             .replace(b'Td', b'Td 3 Tr')
            doc.update_stream(xref, stream)


    def _clip_page(self, bbox:fitz.Rect=None, zoom:float=3.0):
        '''Clip page pixmap (without text) according to ``bbox``.

        Args:
            page (fitz.Page): pdf page to extract.
            bbox (fitz.Rect, optional): Target area to clip. Defaults to None, i.e. entire page.
            zoom (float, optional): Improve resolution by this rate. Defaults to 3.0.

        Returns:
            fitz.Pixmap: The extracted pixmap.
        '''        
        # hide text 
        self._hide_page_text()
        
        # improve resolution
        # - https://pymupdf.readthedocs.io/en/latest/faq.html#how-to-increase-image-resolution
        # - https://github.com/pymupdf/PyMuPDF/issues/181
        bbox = self._page.rect if bbox is None else bbox & self._page.rect
        return self._page.getPixmap(clip=bbox, matrix=fitz.Matrix(zoom, zoom)) # type: fitz.Pixmap

   
    @staticmethod
    def _recover_pixmap(doc:fitz.Document, item:list):
        """Restore pixmap with soft mask considered.
        
        References:

            * https://pymupdf.readthedocs.io/en/latest/document.html#Document.getPageImageList        
            * https://pymupdf.readthedocs.io/en/latest/faq.html#how-to-handle-stencil-masks
            * https://github.com/pymupdf/PyMuPDF/issues/670

        Args:
            doc (fitz.Document): pdf document.
            item (list): image instance of ``page.getImageList()``.

        Returns:
            fitz.Pixmap: Recovered pixmap with soft mask considered.
        """
        # data structure of `item`:
        # (xref, smask, width, height, bpc, colorspace, ...)
        x = item[0]  # xref of PDF image
        s = item[1]  # xref of its /SMask

        # base image
        pix = fitz.Pixmap(doc, x)

        # reconstruct the alpha channel with the smask if exists
        if s > 0:        
            # copy of base image, with an alpha channel added
            pix = fitz.Pixmap(pix, 1)  
            
            # create pixmap of the /SMask entry
            ba = bytearray(fitz.Pixmap(doc, s).samples)
            for i in range(len(ba)):
                if ba[i] > 0: ba[i] = 255
            pix.setAlpha(ba)

        # we may need to adjust something for CMYK pixmaps here -> 
        # recreate pixmap in RGB color space if necessary
        # NOTE: pix.colorspace may be None for images with alpha channel values only
        if pix.colorspace and not pix.colorspace.name in (fitz.csGRAY.name, fitz.csRGB.name):
            pix = fitz.Pixmap(fitz.csRGB, pix)

        return pix


# -*- coding: utf-8 -*-
from enum import Enum
import random
from collections import deque
from collections.abc import Iterable
from fitz.utils import getColorList, getColorInfoList


class BlockType(Enum):
    '''Block types.'''
    UNDEFINED = -1
    TEXT = 0
    IMAGE = 1
    LATTICE_TABLE = 2
    STREAM_TABLE = 3
    FLOAT_IMAGE = 4


class RectType(Enum):
    '''Shape type in context.'''
    HIGHLIGHT = 1
    UNDERLINE = 1<<1
    STRIKE = 1<<2
    HYPERLINK = 1<<3
    BORDER = 1<<4
    SHADING = 1<<5


class TextDirection(Enum):
    '''Text direction.

    * LEFT_RIGHT: from left to right within a line, and lines go from top to bottom
    * BOTTOM_TOP: from bottom to top within a line, and lines go from left to right
    '''
    IGNORE     = -1
    LEFT_RIGHT = 0
    BOTTOM_TOP = 1


class TextAlignment(Enum):
    '''Text alignment.

    .. note::
        The difference between ``NONE`` and ``UNKNOWN``: 

        * NONE: none of left/right/center align -> need TAB stop
        * UNKNOWN: can't decide, e.g. single line only
    '''
    NONE    = -1
    UNKNOWN = 0 
    LEFT    = 1
    CENTER  = 2
    RIGHT   = 3
    JUSTIFY = 4


class IText:
    '''Text related interface considering text direction.'''
    @property
    def text_direction(self):
        '''Text direction is from left to right by default.'''
        return TextDirection.LEFT_RIGHT

    @property
    def is_horizontal_text(self):
        '''Check whether text direction is from left to right.'''
        return self.text_direction == TextDirection.LEFT_RIGHT

    @property
    def is_vertical_text(self):
        '''Check whether text direction is from bottom to top.'''
        return self.text_direction == TextDirection.BOTTOM_TOP


class lazyproperty:
    '''Calculate only once and cache property value.'''
    def __init__(self, func):
        self.func = func

    def __get__(self, instance, cls):
        if instance is None:
            return self
        else:
            value = self.func(instance)
            setattr(instance, self.func.__name__, value)
            return value


# -------------------------
# methods
# -------------------------
def is_number(str_number):
    '''Whether can be converted to a float.'''
    try:
        float(str_number)
    except ValueError:
        return False
    else:
        return True


def flatten(items, klass):
    '''Yield items from any nested iterable.'''
    for item in items:
        if isinstance(item, Iterable) and not isinstance(item, klass):
            yield from flatten(item, klass)
        else:
            yield item


# -------------------------
# color methods
# -------------------------
def rgb_component_from_name(name:str=''):
    '''Get a named RGB color (or random color) from fitz predefined colors, e.g. 'red' -> (1.0,0.0,0.0).'''
    # get color index
    if name and name.upper() in getColorList():
        pos = getColorList().index(name.upper())
    else:
        pos = random.randint(0, len(getColorList())-1)
        
    c = getColorInfoList()[pos]
    return (c[1] / 255.0, c[2] / 255.0, c[3] / 255.0)


def rgb_component(srgb:int):
    '''srgb value to R,G,B components, e.g. 16711680 -> (255, 0, 0).
    
    Equal to PyMuPDF built-in method::

        [int(255*x) for x in fitz.sRGB_to_pdf(x)]
    '''
    # decimal to hex: 0x...
    s = hex(srgb)[2:].zfill(6)
    return [int(s[i:i+2], 16) for i in [0, 2, 4]]


def rgb_to_value(rgb:list):
    '''RGB components to decimal value, e.g. (1,0,0) -> 16711680.'''
    res = 0
    for (i,x) in enumerate(rgb):
        res += int(x*(16**2-1)) * 16**(4-2*i)
    return int(res)


def cmyk_to_rgb(c:float, m:float, y:float, k:float, cmyk_scale:float=100):
    '''CMYK components to GRB value.'''
    r = (1.0 - c / float(cmyk_scale)) * (1.0 - k / float(cmyk_scale))
    g = (1.0 - m / float(cmyk_scale)) * (1.0 - k / float(cmyk_scale))
    b = (1.0 - y / float(cmyk_scale)) * (1.0 - k / float(cmyk_scale))
    res = rgb_to_value([r, g, b]) # type: int
    return res


def rgb_value(components:list):
    '''Gray/RGB/CMYK mode components to color value.'''
    num = len(components)
    # CMYK mode
    if num==4:
        c, m, y, k = map(float, components)
        color = cmyk_to_rgb(c, m, y, k, cmyk_scale=1.0)
    # RGB mode
    elif num==3:
        r, g, b = map(float, components)
        color = rgb_to_value([r, g, b])
    # gray mode
    elif num==1:
        g = float(components[0])
        color = rgb_to_value([g,g,g])    
    else:
        color = 0

    return color


# -------------------------
# pdf plot
# -------------------------
def new_page(doc, width:float, height:float, title:str):
    '''Insert a new page with given title.
    
    Args:
        doc (fitz.Document): pdf document object.
        width (float): Page width.
        height (float): Page height.
        title (str): Page title shown in page.
    '''
    # insert a new page
    page = doc.newPage(width=width, height=height)    

    # plot title at the top-left corner
    gray = rgb_component_from_name('gray')
    page.insertText((5, 16), title, color=gray, fontsize=15)
    
    return page


def debug_plot(title:str, show=True):
    '''Plot the returned objects of inner function.
    
    Args:
        title (str): Page title.
        show (bool, optional): Don't plot if show==False. Default to True.
    
    .. note::
        Prerequisite of the inner function: 
            - the first argument is a :py:class:`~pdf2docxnogui.page.BasePage` instance.
            - the last argument is configuration parameters in ``dict`` type.
    '''
    def wrapper(func):
        def inner(*args, **kwargs):
            # execute function
            objects = func(*args, **kwargs)

            # check if plot page
            page = args[0] # BasePage object
            debug = kwargs.get('debug', False)
            doc = kwargs.get('debug_doc', None)
            filename = kwargs.get('debug_filename', None)

            if show and objects and debug and doc is not None:                
                # create a new page
                debug_page = new_page(doc, page.width, page.height, title)
                # plot objects, e.g. text blocks, shapes, tables...
                objects.plot(debug_page)
                doc.save(filename)

            return objects
        return inner
    return wrapper


# -------------------------------------------------------------------------------------------
# Implementation of solving Rectangle-Intersection Problem according algorithm proposed in
# paper titled "A Rectangle-Intersection Algorithm with Limited Resource Requirements".
# https://ieeexplore.ieee.org/document/5578313
# - Input
# The rectangle is represented by its corner points, (x0, y0, x1, y1)
# - Output
# The output is an Adjacent List of each rect, which could be used to initialize a GRAPH.
# -------------------------------------------------------------------------------------------
# procedure report(S, n)
# 1 Let V be the list of x-coordinates of the 2n vertical edges in S sorted in non-decreasing order.
# 2 Let H be the list of n y-intervals corresponding to the bottom and top y-coordinates of each rectangle.
# 3 Sort the elements of H in non-decreasing order by their bottom y-coordinates.
# 4 Call procedure detect(V, H, 2n).

def solve_rects_intersection(V:list, num:int, index_groups:list):
    '''Implementation of solving Rectangle-Intersection Problem.

    Performance::

        O(nlog n + k) time and O(n) space, where k is the count of intersection pairs.

    Args:
        V (list): Rectangle-related x-edges data, [(index, Rect, x), (...), ...].
        num (int): Count of V instances, equal to len(V).
        index_groups (list): Target adjacent list for connectivity between rects.
    
    Procedure ``detect(V, H, m)``::
    
        if m < 2 then return else
        - let V1 be the first ⌊m/2⌋ and let V2 be the rest of the vertical edges in V in the sorted order;
        - let S11 and S22 be the set of rectangles represented only in V1 and V2 but not spanning V2 and V1, respectively;
        - let S12 be the set of rectangles represented only in V1 and spanning V2; 
        - let S21 be the set of rectangles represented only in V2 and spanning V1
        - let H1 and H2 be the list of y-intervals corresponding to the elements of V1 and V2 respectively
        - stab(S12, S22); stab(S21, S11); stab(S12, S21)
        - detect(V1, H1, ⌊m/2⌋); detect(V2, H2, m − ⌊m/2⌋)
    '''
    if num < 2: return
    
    # start/end points of left/right intervals
    center_pos = int(num/2.0)
    X0, X, X1 = V[0][-1], V[center_pos-1][-1], V[-1][-1] 

    # split into two groups
    left = V[0:center_pos]
    right = V[center_pos:]

    # filter rects according to their position to each intervals
    S11 = list(filter( lambda item: item[1][2]<=X, left ))
    S12 = list(filter( lambda item: item[1][2]>=X1, left ))
    S22 = list(filter( lambda item: item[1][0]>X, right ))
    S21 = list(filter( lambda item: item[1][0]<=X0, right ))
    
    # intersection in x-direction is fullfilled, so check y-direction further
    _stab(S12, S22, index_groups)
    _stab(S21, S11, index_groups)
    _stab(S12, S21, index_groups)

    # recursive process
    solve_rects_intersection(left,  center_pos,     index_groups)
    solve_rects_intersection(right, num-center_pos, index_groups)


def _stab(S1:list, S2:list, index_groups:list):
    '''Check interval intersection in y-direction.
    
    Procedure ``stab(A, B)``::
        i := 1; j := 1
        while i ≤ |A| and j ≤ |B|
            if ai.y0 < bj.y0 then
            k := j
            while k ≤ |B| and bk.y0 < ai.y1
                reportPair(air, bks)
                k := k + 1
            i := i + 1
            else
            k := i
            while k ≤ |A| and ak.y0 < bj.y1
                reportPair(bjs, akr)
                k := k + 1
            j := j + 1
    '''
    if not S1 or not S2: return

    # sort
    S1.sort(key=lambda item: item[1][1])
    S2.sort(key=lambda item: item[1][1])

    i, j = 0, 0
    while i<len(S1) and j<len(S2):
        m, a, _ = S1[i]
        n, b, _ = S2[j]
        if a[1] < b[1]:
            k = j
            while k<len(S2) and S2[k][1][1] < a[3]:
                _report_pair(int(m/2), int(S2[k][0]/2), index_groups)
                k += 1
            i += 1
        else:
            k = i
            while k<len(S1) and S1[k][1][1] < b[3]:
                _report_pair(int(S1[k][0]/2), int(n/2), index_groups)
                k += 1
            j += 1


def _report_pair(i:int, j:int, index_groups:list):
    '''add pair (i,j) to adjacent list.'''
    index_groups[i].add(j)
    index_groups[j].add(i)


def graph_bfs(graph): 
    '''Breadth First Search graph (may be disconnected graph).
    
    Args:
        graph (list): GRAPH represented by adjacent list, [set(1,2,3), set(...), ...]
    
    Returns:
        list: A list of connected components
    '''
    # search graph
    # NOTE: generally a disconnected graph
    counted_indexes = set() # type: set[int]
    groups = []
    for i in range(len(graph)):
        if i in counted_indexes: continue
        # connected component starts...
        indexes = set(_graph_bfs_from_node(graph, i))
        groups.append(indexes)
        counted_indexes.update(indexes)

    return groups


def _graph_bfs_from_node(graph, start):
    '''Breadth First Search connected graph with start node.
    
    Args:
        graph (list): GRAPH represented by adjacent list, [set(1,2,3), set(...), ...].
        start (int): Index of any start vertex.
    '''
    search_queue = deque()    
    searched = set()

    search_queue.append(start)
    while search_queue:
        cur_node = search_queue.popleft()
        if cur_node in searched: continue
        yield cur_node
        searched.add(cur_node)
        for node in graph[cur_node]:
            search_queue.append(node)


# -*- coding: utf-8 -*-

'''A group of ``TextBlock``, ``ImageBlock`` or ``TableBlock``.
'''

from docx.shared import Pt
from ..common import constants
from ..common.Collection import ElementCollection
from ..common.share import BlockType, rgb_value
from ..common.Block import Block
from ..common.docx import reset_paragraph_format, delete_paragraph
from ..text.TextBlock import TextBlock
from ..text.Line import Line
from ..text.Lines import Lines
from ..table.Cell import Cell
from ..image.ImageBlock import ImageBlock
from ..table.TableBlock import TableBlock



class Blocks(ElementCollection):
    '''Block collections.'''
    def __init__(self, instances:list=None, parent=None):
        ''' A collection of TextBlock and TableBlock instances.'''
        super().__init__(instances, parent)
        self._floating_image_blocks = []


    def _update_bbox(self, block:Block):
        '''Override. The parent is ``Layout``, which is not necessary to update its bbox. 
        So, do nothing but required here.
        '''
        pass


    @property
    def floating_image_blocks(self):
        return self._floating_image_blocks


    @property
    def lattice_table_blocks(self):
        '''Get lattice table blocks contained in this Collection.'''
        return list(filter(
            lambda block: block.is_lattice_table_block, self._instances))


    @property
    def stream_table_blocks(self):
        '''Get stream table blocks contained in this Collection.'''
        return list(filter(
            lambda block: block.is_stream_table_block, self._instances))


    @property
    def table_blocks(self):
        '''Get table blocks contained in this Collection.'''
        return list(filter(
            lambda block: block.is_table_block, self._instances))


    @property
    def inline_image_blocks(self):
        '''Get inline image blocks contained in this Collection.'''
        return list(filter(
            lambda block: block.is_inline_image_block, self._instances))


    @property
    def text_blocks(self):
        '''Get text/image blocks contained in this Collection.'''
        return list(filter(
            lambda block: block.is_text_image_block, self._instances))


    def restore(self, raws:list):
        '''Clean current instances and restore them from source dict.
        ImageBlock is converted to ImageSpan contained in TextBlock.

        Args:
            raws (list): A list of raw dicts representing text/image/table blocks.

        Returns:
            Blocks: self
        '''
        self.reset()  # clean current instances
        for raw_block in raws:
            block_type = raw_block.get('type', -1) # type: int
            
            # inline image block -> text block
            if block_type == BlockType.IMAGE.value:
                block = ImageBlock(raw_block).to_text_block()
            
            # text block
            elif block_type == BlockType.TEXT.value:
                block = TextBlock(raw_block)
            
            # table block
            elif block_type in (BlockType.LATTICE_TABLE.value, BlockType.STREAM_TABLE.value):
                block = TableBlock(raw_block)
            
            else:
                block = None            
            
            # add to list
            self.append(block)
        
        return self


    def clean_up(self, delete_end_line_hyphen:bool, float_image_ignorable_gap:float):
        """Clean up blocks in page level.

        * remove blocks out of page
        * remove transformed text: text direction is not (1, 0) or (0, -1)
        * remove empty blocks

        Args:
            float_image_ignorable_gap (float): Regarded as float image if the intersection exceeds this value.

        .. note::
            This method works ONLY for layout initialized from raw dict extracted by ``page.getText()``.
            Under this circumstance, it only exists text blocks since all raw image blocks are converted to 
            text blocks.
        """
        if not self._instances: return
        
        page_bbox = self.parent.working_bbox
        f = lambda block:   block.bbox.intersects(page_bbox) and \
                            not block.white_space_only and (
                            block.is_horizontal_text or block.is_vertical_text)
        self.reset(filter(f, self._instances))

        # sort
        self.strip(delete_end_line_hyphen) \
            .sort_in_reading_order() \
            .identify_floating_images(float_image_ignorable_gap)
   

    def strip(self, delete_end_line_hyphen:bool):
        '''Remove redundant blanks exist in text block lines. These redundant blanks may affect bbox of text block.
        '''
        for block in self._instances: block.strip(delete_end_line_hyphen)
        return self


    def identify_floating_images(self, float_image_ignorable_gap:float):
        """Identify floating image lines and convert to ImageBlock.

        Args:
            float_image_ignorable_gap (float): Regarded as float image if the intersection exceeds this value.
        """
        # group lines by connectivity
        lines = Lines()
        for block in self._instances:
            lines.extend(block.lines)
        groups = lines.group_by_connectivity(dx=-float_image_ignorable_gap, dy=-float_image_ignorable_gap)
        
        # identify floating objects
        for group in filter(lambda group: len(group)>1, groups):
            for line in group:
                block = line.parent
                # consider image block only (converted to text block)
                if not block.is_inline_image_block: continue

                float_image = ImageBlock().from_text_block(block)
                float_image.set_float_image_block()
                self._floating_image_blocks.append(float_image)

                # remove the original image block from flow layout
                block.update_bbox((0,0,0,0))

        return self


    def assign_to_tables(self, tables:list):
        """Add Text/Image/table block lines to associated cells of given tables.

        Args:
            tables (list): A list of TableBlock instances.
        """        
        if not tables: return

        # assign blocks to table region        
        blocks_in_tables = [[] for _ in tables] # type: list[list[Block]]
        blocks = []   # type: list[Block]
        for block in self._instances:
            self._assign_block_to_tables(block, tables, blocks_in_tables, blocks)

        # assign blocks to associated cells
        for table, blocks_in_table in zip(tables, blocks_in_tables):
            # no contents for this table
            if not blocks_in_table: continue
            table.assign_blocks(blocks_in_table)

        # sort in natural reading order and update layout blocks
        blocks.extend(tables)
        self.reset(blocks).sort_in_reading_order()        


    def collect_stream_lines(self, potential_shadings:list, 
            float_layout_tolerance:float, 
            line_separate_threshold:float):
        '''Collect elements in Line level, which may contained in a stream table region.

        * Lines in text block
        * The entire bbox of table block
        
        Table may exist on the following conditions:

        * (a) lines in potential shading -> determined by shapes
        * (b) lines in blocks are not connected sequently -> determined by current block only
        * (c) multi-blocks are in a same row (horizontally aligned) -> determined by two adjacent blocks
        
        Args:
            potential_shadings (list): a group of shapes representing potential cell shading
        
        Returns:
            list: A list of Lines. Each group of Lines represents a potential table.
        
        .. note::
            ``PyMuPDF`` may group multi-lines in a row as a text block while each line belongs to different
            cell. So, it's required to deep into line level.
        '''
        # get sub-lines from block
        def sub_lines(block):
            if block.is_text_image_block:
                return block.lines
            elif block.is_table_block:
                return [Line().update_bbox(block.outer_bbox)]
            else:
                return [Line().update_bbox(block.bbox)]
        
        # exclude potential shading in white bg-color
        shadings_exclude_white = list(filter(
            lambda shape: shape.color != rgb_value([1,1,1]), potential_shadings
        ))
        
        # check block by block
        res = [] # type: list[Lines]
        j = 0
        table_lines = Lines() # potential text lines in a table
        start_new_row = True
        num_blocks, num_shadings = len(self._instances), len(shadings_exclude_white)
        for i in range(num_blocks):

            block = self._instances[i]
            next_block = self._instances[i+1] if i<num_blocks-1 else Block()

            table_end = False

            # (a) block in potential shading?
            if j < num_shadings:
                shading = shadings_exclude_white[j]
                if not shading.is_determined and shading.contains(block, threshold=constants.FACTOR_MOST):
                    table_lines.extend(sub_lines(block))
                    start_new_row = False
                
                # move to next shading
                elif next_block.bbox.y0 > shading.bbox.y1:
                    j += 1
            
            # (b) add lines when current block is not flow layout
            if start_new_row and not block.is_flow_layout(float_layout_tolerance, line_separate_threshold): 
                table_lines.extend(sub_lines(block))  # deep into line level
                
                # update line status
                start_new_row = False            

            # (c) multi-blocks are in a same row: check layout with next block?
            # yes, add both current and next blocks
            if block.horizontally_align_with(next_block, factor=0):
                # if it's start of new table row: add the first block
                if start_new_row: table_lines.extend(sub_lines(block))
                
                # add next block
                table_lines.extend(sub_lines(next_block))

                # update line status
                start_new_row = False

            # no, consider to start a new row
            else:
                # current block is flow layout, and starts in a new row => 
                # close the table before this block
                # ===   ===   ===  <- pre-table
                # ===current block=== <- new row start
                # ===next-block===
                if start_new_row: 
                    table_end = True

                # if the vertical distance is large enough => 
                # close the table before next block
                # ===   ===  ===current-block===
                # significant vertical distance here
                # ===next-block=== ==== <- no matter next block is flow layout or not
                if next_block.bbox.y0-block.bbox.y1>=50:
                    table_end = True

                # update line status            
                start_new_row = True

            # NOTE: close table detecting manually if last block
            if i==num_blocks-1: table_end = True

            # end of current table
            if table_lines and table_end: 
                res.append(table_lines)                
                table_lines = Lines() # reset table_blocks

        return res


    def parse_spacing(self, *args):
        '''Calculate external and internal space for text blocks:

        - vertical distance between blocks, i.e. paragraph before/after spacing
        - horizontal distance to left/right border, i.e. paragraph left/right indent
        - vertical distance between lines, i.e. paragraph line spacing
        '''
        if not self._instances: return
        self._parse_block_horizontal_spacing(*args)
        self._parse_block_vertical_spacing()
        self._parse_line_spacing()


    def join_horizontally(self, 
                text_direction:bool, 
                line_overlap_threshold:float, 
                line_merging_threshold:float):
        '''Join lines in horizontally aligned blocks into new TextBlock.

        This function converts potential float layout into flow layout, e.g. 
        reposition inline images, so that make rebuilding such layout in docx possible.
        
        Args:
            text_direction (bool): Whether consider text direction. 
                                   Detect text direction based on line direction if True, 
                                   otherwise use default direction, i.e. from left to right.
        '''
        # get horizontally aligned blocks group by group
        fun = lambda a,b: a.horizontally_align_with(b, factor=0.0, text_direction=text_direction)
        groups = self.group(fun)
        
        # merge text blocks in each group
        blocks = []
        for blocks_collection in groups:
            # combine all lines into a TextBlock
            final_block = TextBlock()
            for block in blocks_collection:
                if not block.is_text_image_block: continue
                final_block.add(block.lines) # keep empty line, may help to identify cell shading

            # merge lines/spans contained in this text block
            # NOTE:            
            # Lines in text block must have same property, e.g. height, vertical distance, 
            # because average line height is used when create docx. However, the contained lines 
            # may be not reasonable after this step. So, this is just a pre-processing step focusing 
            # on processing lines in horizontal direction, e.g. merging inline image to its text line.
            # Further steps, e.g. split back to original blocks, must be applied before further parsing.
            final_block.lines.join(line_overlap_threshold, line_merging_threshold)

            blocks.append(final_block)
        
        # add table blocks
        blocks.extend(self.table_blocks)
        self.reset(blocks)
        

        return self


    def join_vertically_by_space(self, block_merging_threshold):
        '''Merge adjacent blocks in vertical direction because blocks belonging to 
        same paragraph might be split by ``PyMuPDF`` unreasonably.
        
        The splitting criterion is that when the distance between blocks:
        
        * is smaller than average line distance when multi-lines; or
        * is smaller than a threshold * block height when single line.

        .. note::
            Considered only normal reading direction, from left to right, from top
            to bottom.
        '''
        blocks = [] # type: list[TextBlock]
        ref = None # type: TextBlock

        # check adjacent two text blocks
        for block in self._instances:
            merged = False

            # add block if previous isn't a text block
            if ref is None or not ref.is_text_image_block:
                blocks.append(block)
            
            # add block if this isn't a text block
            elif not block.is_text_image_block:
                blocks.append(block)
            
            # check two adjacent text blocks
            else:
                # block gap
                idx = 1 if ref.is_horizontal_text else 0                
                gap_block = block.bbox[idx] - ref.bbox[idx+2]

                # lines gap
                gap_line1, gap_line2 = ref.average_row_gap, block.average_row_gap

                # single line blocks
                if gap_line1==gap_line2==None:
                    # block height
                    h1 = ref.bbox[idx+2]-ref.bbox[idx]
                    h2 = block.bbox[idx+2]-block.bbox[idx]
                    merged = abs(gap_block-block_merging_threshold*min(h1, h2))<=constants.TINY_DIST
                
                # multi-lines block
                else:                    
                    gap_line = gap_line1 if not gap_line1 is None else gap_line2
                    merged = abs(gap_block-gap_line) <= constants.TINY_DIST
                
                if merged:
                    ref.add(block.lines)
                else:
                    blocks.append(block)

            # NOTE: update ref block only no merging happens
            if not merged: ref = block
       
        self.reset(blocks)


    def split_back(self, *args):
        '''Split the joined lines back to original text block if possible.

        With preceding joining step, current text block may contain lines coming from various original blocks.
        Considering that different text block may have different line properties, e.g. height, spacing, 
        this function is to split them back to original text block. 

        .. note::
            Don't split block if the splitting breaks flow layout, e.g. two blocks (i.e. two paragraphs in docx) 
            in same row.
        '''
        blocks = [] # type: list[TextBlock]
        lines = Lines()
        # collect lines for further step, or table block directly
        for block in self._instances:
            if block.is_text_image_block and block.is_flow_layout(*args):
                lines.extend([line for line in block.lines if not line.white_space_only]) # filter empty line
            else:
                blocks.append(block)
        
        # regroup lines
        for group_lines in lines.split_back():
            text_block = TextBlock()
            text_block.lines.reset(group_lines)
            blocks.append(text_block)
       
        self.reset(blocks).sort_in_reading_order()


    def split_vertically_by_text(self, line_break_free_space_ratio:float, new_paragraph_free_space_ratio:float):
        '''Split text block into separate paragraph based on punctuation of sentense.

        .. note::
            Considered only normal reading direction, from left to right, from top
            to bottom.
        '''
        blocks = [] # type: list[TextBlock]
        for block in self._instances:

            # add block if this isn't a text block
            if not block.is_text_block: 
                blocks.append(block)
                continue
            
            # add split blocks if necessary
            lines_list = block.lines.split_vertically_by_text(line_break_free_space_ratio, 
                                                                new_paragraph_free_space_ratio)
            if len(lines_list)==1:
                blocks.append(block)
            else:
                for lines in lines_list:
                    text_block = TextBlock()
                    text_block.add(lines)
                    blocks.append(text_block)

        self.reset(blocks)


    def parse_text_format(self, rects):
        '''Parse text format with style represented by stroke/fill shapes.
        
        Args:
            rects (Shapes): Potential styles applied on blocks.
        '''
        # parse text block style one by one
        for block in filter(lambda e: e.is_text_block, self._instances): 
            block.parse_text_format(rects)


    def make_docx(self, doc):
        '''Create page based on parsed block structure. 
        
        Args:
            doc (Document, _Cell): The container to make docx content.
        '''
        def make_table(table_block, pre_table):
            # create dummy paragraph if table before space is set
            # - a minimum line height of paragraph is 0.7pt, so ignore before space if less than this value
            # - but tow adjacent tables will be combined automatically, so adding a minimum dummy paragraph is required
            if table_block.before_space>=constants.MIN_LINE_SPACING or pre_table:
                h = int(10*table_block.before_space)/10.0 # round(x,1), but to lower bound
                p = doc.add_paragraph()
                reset_paragraph_format(p, line_spacing=Pt(h))

            # new table            
            table = doc.add_table(rows=table_block.num_rows, cols=table_block.num_cols)
            table.autofit = False
            table.allow_autofit  = False
            table_block.make_docx(table)

        pre_table = False
        cell_layout = isinstance(self.parent, Cell)
        for block in self._instances:
            # make paragraphs
            if block.is_text_image_block:                
                # new paragraph
                p = doc.add_paragraph()
                block.make_docx(p)

                pre_table = False # mark block type
            
            # make table
            elif block.is_table_block:
                make_table(block, pre_table)
                pre_table = True # mark block type

                # NOTE: within a cell, there is always an empty paragraph after table,
                # so, delete it right here.
                # https://github.com/dothinking/pdf2docxnogui/issues/76 
                if cell_layout:
                    delete_paragraph(doc.paragraphs[-1])
       
        # NOTE: If a table is at the end of a page, a new paragraph will be automatically 
        # added by the rending engine, e.g. MS Word, which resulting in an unexpected
        # page break. The solution is to never put a table at the end of a page, so add
        # an empty paragraph and reset its format, particularly line spacing, when a table
        # is created.
        for block in self._instances[::-1]:
            # ignore float image block
            if block.is_float_image_block: continue

            # nothing to do if not end with table block
            if not block.is_table_block: break

            # otherwise, add a small paragraph
            p = doc.add_paragraph()
            reset_paragraph_format(p, Pt(constants.MIN_LINE_SPACING)) # a small line height

  
    def plot(self, page):
        '''Plot blocks in PDF page for debug purpose.'''
        for block in self._instances: block.plot(page)                


    @staticmethod
    def _assign_block_to_tables(block:Block, tables:list, blocks_in_tables:list, blocks:list):
        '''Collect blocks contained in table region ``blocks_in_tables`` and rest text blocks in ``blocks``.'''
        # lines in block for further check if necessary
        lines = block.lines if block.is_text_image_block else Lines()

        # collect blocks contained in table region
        # NOTE: there is a probability that only a part of a text block is contained in table region, 
        # while the rest is in normal block region.
        for table, blocks_in_table in zip(tables, blocks_in_tables):

            # fully contained in one table
            if table.bbox.contains(block.bbox):
                blocks_in_table.append(block)
                break
            
            # not possible in current table, then check next table
            elif not table.bbox.intersects(block.bbox): continue
            
            # deep into line level for text block
            elif block.is_text_image_block:
                table_lines, not_table_lines = lines.split_with_intersection(table.bbox, constants.FACTOR_MOST)

                # add lines to table
                text_block = TextBlock()
                text_block.add(table_lines)
                blocks_in_table.append(text_block)

                # lines not in table for further check
                if not_table_lines:
                    lines = not_table_lines
                else:
                    break # no more lines
        
        # Now, this block (or part of it) belongs to previous layout
        else:
            if block.is_table_block:
                blocks.append(block)
            else:
                text_block = TextBlock()
                text_block.add(lines)
                blocks.append(text_block)


    def _parse_block_horizontal_spacing(self, *args):
        '''Calculate external horizontal space for text blocks, i.e. alignment mode and left spacing 
        for paragraph in docx:
        
            - horizontal block -> take left boundary as reference
            - vertical block   -> take bottom boundary as reference
        '''
        # bbox of blocks
        # - page level, e.g. blocks in top layout
        # - table level, e.g. blocks in table cell
        bbox = self.parent.working_bbox

        for block in self._instances:
            block.parse_horizontal_spacing(bbox, *args)


    def _parse_block_vertical_spacing(self):
        '''Calculate external vertical space for text blocks, i.e. before/after space in docx.
        
        The vertical spacing is determined by the vertical distance to previous block.
        For the first block, the reference position is top margin.

        It's easy to set before-space or after-space for a paragraph with ``python-docx``,
        so, if current block is a paragraph, set before-space for it; if current block is 
        not a paragraph, e.g. a table, set after-space for previous block (generally, 
        previous block should be a paragraph).
        '''
        # bbox of blocks
        # - page level, e.g. blocks in top layout
        # - table level, e.g. blocks in table cell
        bbox = self.parent.working_bbox

        # check text direction for vertical space calculation:
        # - normal reading direction (from left to right)    -> the reference boundary is top border, i.e. bbox[1].
        # - vertical text direction, e.g. from bottom to top -> left border bbox[0] is the reference
        idx = 1 if self.is_horizontal_text else 0

        ref_block = self._instances[0]
        ref_pos = bbox[idx]

        for block in self._instances:

            # NOTE: the table bbox is counted on center-line of outer borders, so a half of top border
            # size should be excluded from the calculated vertical spacing
            if block.is_table_block:
                dw = block[0][0].border_width[0] / 2.0 # use top border of the first cell
            
            else:
                dw = 0.0

            start_pos = block.bbox[idx] - dw
            para_space = start_pos-ref_pos

            # modify vertical space in case the block is out of bottom boundary
            dy = max(block.bbox[idx+2]-bbox[idx+2], 0.0)
            para_space -= dy
            para_space = max(para_space, 0.0) # ignore negative value

            # ref to current (paragraph): set before-space for paragraph
            if block.is_text_block:
                # spacing before this paragraph
                block.before_space = para_space

            # if ref to current (image): set before-space for paragraph
            elif block.is_inline_image_block:
                block.before_space = para_space

            # ref (paragraph/image) to current: set after-space for ref paragraph        
            elif ref_block.is_text_block or ref_block.is_inline_image_block:
                ref_block.after_space = para_space

            # situation with very low probability, e.g. ref (table) to current (table)
            # we can't set before space for table in docx, but the tricky way is to
            # create an empty paragraph and set paragraph line spacing and before space
            else:
                # let para_space>=1 Pt to accommodate the dummy paragraph if not the first block
                block.before_space = max(para_space, int(block!=self._instances[0])*constants.MINOR_DIST)

            # update reference block        
            ref_block = block
            ref_pos = ref_block.bbox[idx+2] + dw # assume same bottom border with top one

        # NOTE: when a table is at the end of a page, a dummy paragraph with a small line spacing 
        # is added after this table, to avoid unexpected page break. Accordingly, this extra spacing 
        # must be remove in other place, especially the page space is run out.
        # Here, reduce the last row of table.
        block = self._instances[-1]
        if block.is_table_block: block[-1].height -= constants.MINOR_DIST


    def _parse_line_spacing(self):
        '''Calculate internal vertical space for text blocks, i.e. paragraph line spacing in docx.

        .. note::
            Run parsing block vertical spacing in advance.
        '''
        for block in self._instances:
            if block.is_text_block:
                block.parse_relative_line_spacing()



# -*- coding: utf-8 -*-


# -------------------------------------
# basic unit
# -------------------------------------
PT = 1.0   # basic unit
ITP = 72.0 # inch to point

MAJOR_DIST = 5.0 * PT   # significant distance exists between two block lines
MINOR_DIST = 1.0 * PT   # small distance
TINY_DIST  = 0.5 * PT   # very small distance

FACTOR_SAME = 0.99
FACTOR_ALMOST = 0.95
FACTOR_MOST = 0.90
FACTOR_MAJOR = 0.75
FACTOR_A_HALF = 0.5
FACTOR_A_FEW = 0.1
FACTOR_FEW = 0.01


# -------------------------------------
# docx
# -------------------------------------
HIDDEN_W_BORDER = 0.0   # do not show border
MIN_LINE_SPACING = 0.7  # minimum line spacing available in MS word
DEFULT_LINE_SPACING = 1.02

# punctuation implying end of a sentense
SENTENSE_END_PUNC = '.．。?？!！'


# -------------------------------------
# font parameters
# -------------------------------------

# CJK font definition parameters
CJK_CODEPAGE_BITS = {
    "JIS/Japan": 17,
    "Chinese: Simplified chars—PRC and Singapore": 18,
    "Korean Wansung": 19,
    "Chinese: Traditional chars—Taiwan and Hong Kong": 20,
    "Korean Johab": 21
}

CJK_UNICODE_RANGE_BITS = {
    'Hangul Jamo': 28,
    'Hiragana': 49,
    'Katakana': 50,
    'Bopomofo': 51,
    'Hangul Compatibility Jamo': 52,
    'Enclosed CJK Letters And Months': 54,
    'CJK Compatibility': 55,
    'Hangul Syllables': 56,
    'CJK Unified Ideographs': 59,
    'CJK Strokes': 61,
    'Yi Syllables': 83
}

CJK_UNICODE_RANGES = [
    [0x1100, 0x11FF],  # Hangul Jamo
    [0x3040, 0x309F],  # Hiragana
    [0x30A0, 0x30FF],  # Katakana
    [0x31F0, 0x31FF],  # Katakana Phonetic Extensions
    [0x3100, 0x312F],  # Bopomofo
    [0x31A0, 0x31BF],  # Bopomofo Extended (Bopomofo)
    [0x3130, 0x318F],  # Hangul Compatibility Jamo
    [0x3200, 0x32FF],  # Enclosed CJK Letters and Months
    [0x3300, 0x33FF],  # CJK Compatibility
    [0xAC00, 0xD7AF],  # Hangul Syllables
    [0x4E00, 0x9FFF],  # CJK Unified Ideographs
    [0x2E80, 0x2EFF],  # CJK Radicals Supplement (CJK Unified Ideographs)
    [0x2F00, 0x2FDF],  # Kangxi Radicals (CJK Unified Ideographs)
    [0x2FF0, 0x2FFF],  # Ideographic Description Characters (CJK Unified Ideographs)
    [0x3400, 0x4DBF],  # CJK Unified Ideographs Extension A (CJK Unified Ideographs)
    [0x20000, 0x2A6DF],  # CJK Unified Ideographs Extension B (CJK Unified Ideographs)
    [0x3190, 0x319F],  # Kanbun (CJK Unified Ideographs)
    [0x31C0, 0x31EF],  # CJK Strokes
    [0xF900, 0xFAFF],  # CJK Compatibility Ideographs (CJK Strokes)
    [0x2F800, 0x2FA1F],  # CJK Compatibility Ideographs Supplement (CJK Strokes)
    [0xA000, 0xA48F],  # Yi Syllables
    [0xA490, 0xA4CF],  # Yi Radicals
]

DEFAULT_FONT_NAME = 'Arial'

# default font name and line height ratio, e.g. PDF base 14 fonts
# https://appligent.com/what-are-the-base-14-fonts/
DICT_FONT_LINE_HEIGHT = {
    'Agency FB': 1.18,
    'Algerian': 1.32,
    'Angsana New': 1.35,
    'Book Antiqua': 1.24,
    'Arial': 1.15,
    'Arial Narrow': 1.15,
    'Arial Black': 1.41,
    'Arial Rounded MT Bold': 1.16,
    'Bahnschrift': 1.2,
    'Barlow Condensed': 1.35,
    'Baskerville Old Face': 1.14,
    'Bauhaus 93': 1.46,
    'Bell MT': 1.13,
    'Bernard MT Condensed': 1.19,
    'Aharoni': 1.06,
    'Andalus': 1.53,
    'AngsanaUPC': 1.35,
    'Aparajita': 1.18,
    'Arabic Typesetting': 1.15,
    'Arial Monospaced': 1.15,
    'Batang': 1.3,
    'Browallia New': 1.25,
    'DaunPenh': 1.36,
    'David': 1.0,
    'DFKai': 1.3,
    'DilleniaUPC': 1.3,
    'DokChampa': 1.94,
    'Estrangelo Edessa': 1.0,
    'EucrosiaUPC': 1.22,
    'Euphemia': 1.35,
    'FangSong': 1.3,
    'FrankRuehl': 1.0,
    'Gisha': 1.17,
    'Gulim': 1.3,
    'IrisUPC': 1.26,
    'Iskoola Pota': 1.14,
    'JasmineUPC': 1.06,
    'KaiTi': 1.3,
    'Kalinga': 1.58,
    'Kartika': 1.0,
    'Khmer UI': 1.13,
    'KodchiangUPC': 0.98,
    'Kokila': 1.15,
    'Lao UI': 1.33,
    'Latha': 1.66,
    'Leelawadee': 1.2,
    'Levenim MT': 1.4,
    'LilyUPC': 0.95,
    'Mangal': 1.68,
    'Microsoft Uighur': 1.08,
    'Miriam Fixed': 1.0,
    'Miriam': 1.06,
    'MS PMincho': 1.3,
    'Nyala': 1.04,
    'Plantagenet Cherokee': 1.32,
    'Raavi': 1.79,
    'Sakkal Majalla': 1.4,
    'SAPDings': 1.0,
    'Shonar Bangla': 1.29,
    'Shruti': 1.83,
    'SimHei': 1.3,
    'Simplified Arabic Fixed': 1.09,
    'Simplified Arabic': 1.66,
    'Terminal': 1.06,
    'Traditional Arabic': 1.49,
    'Tunga': 1.77,
    'Utsaah': 1.12,
    'Vani': 1.68,
    'Vrinda': 1.02,
    'Yu Gothic Light': 1.43,
    'Bodoni MT': 1.2,
    'Bodoni MT Black': 1.17,
    'Bodoni MT Condensed': 1.18,
    'Bodoni MT Poster Compressed': 1.15,
    'Bookman Old Style': 1.17,
    'Bradley Hand ITC': 1.25,
    'Britannic Bold': 1.11,
    'Berlin Sans FB': 1.1,
    'Berlin Sans FB Demi': 1.13,
    'Broadway': 1.13,
    'BrowalliaUPC': 1.25,
    'Brush Script MT': 1.23,
    'Bookshelf Symbol 7': 1.0,
    'Calibri': 1.22,
    'Californian FB': 1.14,
    'Calisto MT': 1.15,
    'Cambria': 1.17,
    'Candara': 1.22,
    'Cascadia Mono': 1.32,
    'Castellar': 1.21,
    'Caveat': 1.28,
    'Century Schoolbook': 1.2,
    'Centaur': 1.14,
    'Century': 1.2,
    'Chiller': 1.15,
    'Code': 1.0,
    'Colonna MT': 1.06,
    'Comic Sans MS': 1.39,
    'Consolas': 1.17,
    'Constantia': 1.22,
    'Cooper Black': 1.15,
    'Copperplate Gothic Bold': 1.11,
    'Copperplate Gothic Light': 1.1,
    'Corbel': 1.22,
    'Corbel Light': 1.22,
    'Cormorant Infant': 1.22,
    'Courier': 1.04,
    'Courier New': 1.13,
    'Curlz MT': 1.33,
    'Dosis': 1.36,
    'Dubai': 1.69,
    'Ebrima': 1.28,
    'Elephant': 1.29,
    'Engravers MT': 1.16,
    'Eras Bold ITC': 1.16,
    'Eras Demi ITC': 1.15,
    'Eras Light ITC': 1.13,
    'Eras Medium ITC': 1.14,
    'Felix Titling': 1.17,
    'Forte': 1.36,
    'Franklin Gothic': 1.13,
    'Freestyle Script': 1.18,
    'French Script MT': 1.14,
    'Footlight MT Light': 1.06,
    'Gabriola': 1.84,
    'Gadugi': 1.33,
    'Garamond': 1.12,
    'Georgia': 1.14,
    'Gigi': 1.38,
    'Gill Sans MT': 1.16,
    'Gill Sans MT Condensed': 1.2,
    'Gill Sans Ultra Bold Condensed': 1.25,
    'Gill Sans Ultra Bold': 1.25,
    'Gloucester MT Extra Condensed': 1.16,
    'Gill Sans MT Ext Condensed Bold': 1.2,
    'Century Gothic': 1.23,
    'Goudy Old Style': 1.2,
    'Goudy Stout': 1.37,
    'Harlow Solid Italic': 1.26,
    'Harrington': 1.18,
    'Haettenschweiler': 1.07,
    'Helvetica': 1.18,
    'Microsoft Himalaya': 1.0,
    'HoloLens MDL2 Assets': 1.0,
    'High Tower Text': 1.16,
    'Impact': 1.22,
    'Imprint MT Shadow': 1.18,
    'Informal Roman': 1.2,
    'Ink Free': 1.24,
    'Blackadder ITC': 1.28,
    'Edwardian Script ITC': 1.18,
    'Kristen ITC': 1.36,
    'Javanese Text': 2.27,
    'Jokerman': 1.51,
    'Juice ITC': 1.19,
    'Kunstler Script': 1.09,
    'Wide Latin': 1.23,
    'Lato': 1.2,
    'Lucida Bright': 1.18,
    'Lucida Calligraphy': 1.36,
    'Leelawadee UI': 1.33,
    'Leelawadee UI Semilight': 1.33,
    'Lucida Fax': 1.18,
    'Lucida Handwriting': 1.38,
    'Lucida Sans': 1.18,
    'Lucida Sans Typewriter': 1.17,
    'Lucida Console': 1.0,
    'Lucida Sans Unicode': 1.54,
    'Magneto': 1.21,
    'Maiandra GD': 1.2,
    'Malgun Gothic': 1.73,
    'Malgun Gothic Semilight': 1.73,
    'Marlett': 1.0,
    'Matura MT Script Capitals': 1.33,
    'Microsoft Sans Serif': 1.13,
    'Mistral': 1.22,
    'Myanmar Text': 1.9,
    'Modern No. 20': 1.07,
    'Mongolian Baiti': 1.15,
    'Montserrat': 1.38,
    'Microsoft Yi Baiti': 1.3,
    'Monotype Corsiva': 1.12,
    'MT Extra': 1.01,
    'MV Boli': 1.61,
    'Niagara Engraved': 1.07,
    'Niagara Solid': 1.07,
    'Nirmala UI': 1.33,
    'Nirmala UI Semilight': 1.33,
    'Noto Sans': 1.36,
    'Noto Serif': 1.36,
    'Microsoft New Tai Lue': 1.31,
    'OCR A Extended': 1.03,
    'Old English Text MT': 1.22,
    'Onyx': 1.15,
    'Open Sans': 1.36,
    'Oswald': 1.7,
    'MS Outlook': 1.04,
    'Palatino Linotype': 1.35,
    'Palace Script MT': 0.93,
    'Papyrus': 1.57,
    'Parchment': 1.07,
    'Perpetua': 1.15,
    'Perpetua Titling MT': 1.18,
    'Microsoft PhagsPa': 1.52,
    'Playbill': 1.01,
    'Poor Richard': 1.13,
    'Pristina': 1.31,
    'Rage Italic': 1.26,
    'Raleway': 1.17,
    'Ravie': 1.33,
    'MS Reference Sans Serif': 1.58,
    'MS Reference Specialty': 1.23,
    'Roboto': 1.2,
    'Roboto Slab': 1.35,
    'Rockwell Condensed': 1.18,
    'Rockwell': 1.17,
    'Script MT Bold': 1.2,
    'Segoe MDL2 Assets': 1.0,
    'Segoe Print': 1.77,
    'Segoe Script': 1.61,
    'Segoe UI': 1.33,
    'Segoe UI Light': 1.33,
    'Segoe UI Semilight': 1.33,
    'Segoe UI Black': 1.33,
    'Segoe UI Emoji': 1.22,
    'Segoe UI Historic': 1.33,
    'Segoe UI Semibold': 1.33,
    'Segoe UI Symbol': 1.73,
    'Showcard Gothic': 1.24,
    'SimSun': 1.3,
    'LiSu': 1.3,
    'YouYuan': 1.3,
    'STCaiyun': 1.35,
    'STFangsong': 1.69,
    'STHupo': 1.34,
    'STKaiti': 1.69,
    'STLiti': 1.38,
    'STSong': 1.69,
    'STXihei': 1.79,
    'STXingkai': 1.42,
    'STXinwei': 1.35,
    'STZhongsong': 1.72,
    'Snap ITC': 1.29,
    'Source Sans Pro Black': 1.26,
    'Source Sans Pro': 1.26,
    'Source Sans Pro ExtraLight': 1.26,
    'Source Sans Pro Light': 1.26,
    'Source Sans Pro SemiBold': 1.26,
    'Stencil': 1.19,
    'Sylfaen': 1.32,
    'Symbol': 1.23,
    'Tahoma': 1.21,
    'Microsoft Tai Le': 1.27,
    'Tw Cen MT': 1.09,
    'Tw Cen MT Condensed': 1.07,
    'Tw Cen MT Condensed Extra Bold': 1.08,
    'Tempus Sans ITC': 1.3,
    'Times': 1.15,
    'Times New Roman': 1.15,
    'Trebuchet MS': 1.16,
    'Verdana': 1.22,
    'Viner Hand ITC': 1.61,
    'Vivaldi': 1.22,
    'Vladimir Script': 1.21,
    'Webdings': 1.0,
    'Wingdings': 1.11,
    'Wingdings 2': 1.05,
    'Wingdings 3': 1.14,
    'Zilla Slab': 1.2
}


# -*- coding: utf-8 -*-

'''A wrapper of ``fitz.Page`` to do the following work:

* extract source contents
* clean up blocks/shapes, e.g. elements out of page
* calculate page margin
* parse page structure roughly, i.e. section and column

The raw page content extracted with ``PyMuPDF`` API ``page.getText('rawdict')`` 
is described per link https://pymupdf.readthedocs.io/en/latest/textpage.html:: 

    {
        # raw dict
        ----------------------------
        "width" : w,
        "height": h,    
        "blocks": [{...}, {...}, ...],

        # introduced dict
        ----------------------------
        "shapes" : [{...}, {...}, ...]
    }

In addition to the raw layout dict, rectangle shapes are also included.
'''

from collections import defaultdict
from .BasePage import BasePage
from ..layout.Layout import Layout
from ..layout.Section import Section
from ..layout.Column import Column
from ..shape.Shape import Hyperlink
from ..shape.Shapes import Shapes
from ..image.ImagesExtractor import ImagesExtractor
from ..shape.Paths import Paths
from ..font.Fonts import Font, Fonts
from ..text.TextSpan import TextSpan
from ..common.Element import Element
from ..common.share import RectType, debug_plot
from ..common import constants
from ..common.Collection import Collection


class RawPage(BasePage, Layout):
    '''A wrapper of ``fitz.Page`` to extract source contents.'''

    def __init__(self, fitz_page=None):
        ''' Initialize page layout.
        
        Args:
            fitz_page (fitz.Page): Source pdf page.
        '''
        BasePage.__init__(self)
        Layout.__init__(self)

        self.fitz_page = fitz_page

    
    @property
    def text(self):
        '''All extracted text in this page, with images considered as ``<image>``. 
        Should be run after ``restore()`` data.'''
        return '\n'.join([block.text for block in self.blocks])

    @property
    def raw_text(self):
        '''Extracted raw text in current page. Should be run after ``restore()`` data.'''
        return '\n'.join([block.raw_text for block in self.blocks])


    @debug_plot('Source Text Blocks')
    def restore(self, **settings):
        '''Initialize layout extracted with ``PyMuPDF``.'''
        raw_dict = self.extract_raw_dict(**settings)
        super().restore(raw_dict)
        return self.blocks

    
    @debug_plot('Cleaned Shapes')
    def clean_up(self, **settings):
        '''Clean up raw blocks and shapes, e.g. 
        
        * remove negative or duplicated instances,
        * merge text blocks horizontally (preparing for layout parsing)
        * detect semantic type of shapes
        '''
        # clean up blocks first
        self.blocks.clean_up(
            settings['delete_end_line_hyphen'],
            settings['float_image_ignorable_gap'])

        # clean up shapes        
        self.shapes.clean_up(
            settings['max_border_width'],
            settings['shape_min_dimension'])
        
        return self.shapes


    def process_font(self, fonts:Fonts, default_font:Font):      
        '''Update font properties, e.g. font name, font line height ratio, of ``TextSpan``.
        
        Args:
            fonts (Fonts): Fonts used in current document.
        '''
        # get all text span
        spans = []
        for block in self.blocks:
            for line in block.lines:
                spans.extend([span for span in line.spans if isinstance(span, TextSpan)])

        # check and update font name, line height
        for span in spans:
            font = fonts.get(span.font, default_font)
            span.font = font.name
            span.line_height = font.line_height * span.size


    def calculate_margin(self, **settings):
        """Calculate page margin.

        .. note::
            Ensure this method is run right after cleaning up the layout, so the page margin is 
            calculated based on valid layout, and stay constant.
        """
        # Exclude hyperlink from shapes because hyperlink might exist out of page unreasonablely, 
        # while it should always within page since attached to text.
        shapes = Shapes([shape for shape in self.shapes if not isinstance(shape, Hyperlink)])

        # return default margin if no blocks exist
        if not self.blocks and not shapes: return (constants.ITP, ) * 4

        x0, y0, x1, y1 = self.bbox
        u0, v0, u1, v1 = self.blocks.bbox | shapes.bbox

        # margin
        left = max(u0-x0, 0.0)
        right = max(x1-u1-constants.MINOR_DIST, 0.0)
        top = max(v0-y0, 0.0)
        bottom = max(y1-v1, 0.0)

        # reduce calculated top/bottom margin to leave some free space
        top *= settings['page_margin_factor_top']
        bottom *= settings['page_margin_factor_bottom']

        # use normal margin if calculated margin is large enough
        return (
            min(constants.ITP, round(left, 1)), 
            min(constants.ITP, round(right, 1)), 
            min(constants.ITP, round(top, 1)), 
            min(constants.ITP, round(bottom, 1)))


    def parse_section(self, **settings):
        '''Detect and create page sections.

        .. note::
            - Only two-columns Sections are considered for now.
            - Page margin must be parsed before this step.
        '''
        # bbox
        X0, Y0, X1, _ = self.working_bbox        
    
        # collect all blocks and shapes
        elements = Collection()
        elements.extend(self.blocks)
        elements.extend(self.shapes)
        if not elements: return
        
        pre_section = Collection()
        pre_num_col = 1
        y_ref = Y0 # to calculate v-distance between sections
        sections = []

        def create_pre_section(num_col, elements, y_ref):
            # append to pre-pre-section if both single column
            if sections and sections[-1].num_cols==num_col==1:
                column = sections[-1][0] # type: Column
                column.union_bbox(elements)
                column.add_elements(elements)
            # otherwise, create new section
            else:
                section = self._create_section(num_col, elements, (X0, X1), y_ref)
                if section: sections.append(section)

        # check section row by row
        for row in elements.group_by_rows():
            # check column col by col
            cols = row.group_by_columns()
            current_num_col = len(cols)

            # column check:
            # - consider 2-cols only
            # - ignore tiny-width column
            if current_num_col>2 or (
                current_num_col==2 and \
                    min(cols[0].bbox.width, cols[1].bbox.width)<=constants.MAJOR_DIST):
                current_num_col = 1

            # process exception
            x0, y0, x1, y1 = pre_section.bbox
            if pre_num_col==2 and current_num_col==1:
                # current row belongs to left column?
                cols = pre_section.group_by_columns()
                if row.bbox[2] <= cols[0].bbox[2]:
                    current_num_col = 2
                
                # further check 2-cols -> the height
                elif y1-y0<settings['min_section_height']:
                    pre_num_col = 1

            elif pre_num_col==2 and current_num_col==2:
                # current 2-cols not align with pre-section ?
                combine = Collection(pre_section)
                combine.extend(row)
                if len(combine.group_by_columns())==1:
                    current_num_col = 1

            # finalize pre-section if different to current section
            if current_num_col!=pre_num_col:
                # process pre-section
                create_pre_section(pre_num_col, pre_section, y_ref)
                if sections: y_ref = sections[-1][-1].bbox[3]

                # start potential new section                
                pre_section = Collection(row)
                pre_num_col = current_num_col

            # otherwise, append to pre-section
            else:
                pre_section.extend(row)

        # the final section
        create_pre_section(current_num_col, pre_section, y_ref)

        return sections


    def extract_raw_dict(self, **settings):
        '''Extract source data from page by ``PyMuPDF``.'''
        if not self.fitz_page: return {}

        # source blocks
        # NOTE: all these coordinates are relative to un-rotated page
        # https://pymupdf.readthedocs.io/en/latest/page.html#modifying-pages
        raw_layout = self.fitz_page.getText('rawdict')

        # page size: though 'width', 'height' are contained in `raw_dict`, 
        # they are based on un-rotated page. So, update page width/height 
        # to right direction in case page is rotated
        *_, w, h = self.fitz_page.rect # always reflecting page rotation
        raw_layout.update({ 'width' : w, 'height': h })
        self.width, self.height = w, h

        # pre-processing for layout blocks and shapes based on parent page
        self._preprocess_images(raw_layout, **settings)
        self._preprocess_shapes(raw_layout, **settings)
       
        # Element is a base class processing coordinates, so set rotation matrix globally
        Element.set_rotation_matrix(self.fitz_page.rotationMatrix)

        return raw_layout


    def _preprocess_images(self, raw, **settings):
        '''Adjust image blocks. 
        
        Image block extracted by ``page.getText('rawdict')`` doesn't contain alpha channel data,
        so it needs to get page images by ``page.getImageList()`` and then recover them. However, 
        ``Page.getImageList()`` contains each image only once, while ``page.getText('rawdict')`` 
        generates image blocks for every image location, whether or not there are any duplicates. 
        See PyMuPDF doc:

        https://pymupdf.readthedocs.io/en/latest/textpage.html#dictionary-structure-of-extractdict-and-extractrawdict
            
        Above all, a compromise:

        * Get image contents with ``page.getImageList()`` -> ensure correct images
        * Get image location with ``page.getText('rawdict')`` -> ensure correct locations
        '''
        # recover image blocks
        recovered_images = ImagesExtractor(self.fitz_page). \
                                extract_images(settings['clip_image_res_ratio'])

        # group original image blocks by image contents
        image_blocks_group = defaultdict(list)
        for block in raw['blocks']:
            if block['type'] != 1: continue
            block['type'] = -1 # "delete" it temporally
            image_blocks_group[hash(block['image'])].append(block)
        
        def same_images(img, img_list):
            bbox = list(map(round, img['bbox']))
            for _img in img_list:
                if list(map(round, _img['bbox']))==bbox: return True
            return False

        # An example to show complicated things here:
        # - images in `page.getImageList`: [a, b, c]
        # - images in `page.getText`     : [a1, a2, b, d]
        # (1) a -> a1, a2: an image in page maps to multi-instances in raw dict
        # (2) c: an image in page may not exist in raw dict -> so, add it
        # (3) d: an image in raw dict may not exist in page -> so, delete it
        for image in recovered_images:
            for k, image_blocks in image_blocks_group.items():
                if not same_images(image, image_blocks): continue
                for image_block in image_blocks:
                    image_block['type'] = 1 # add it back
                    image_block['image'] = image['image']
                break

            # an image outside the page is not counted in page.getText(), so let's add it here
            else:
                raw['blocks'].append(image)


    @debug_plot('Source Paths')
    def _preprocess_shapes(self, raw, **settings):
        '''Identify iso-oriented paths and convert vector graphic paths to pixmap.'''
        # extract paths ed by `page.getDrawings()`
        raw_paths = self.fitz_page.getDrawings()

        # iso-oriented paths to shapes
        paths = Paths(parent=self).restore(raw_paths)
        shapes, iso_areas, exist_svg = paths.to_shapes(settings['curve_path_ratio'])
        raw['shapes'] = shapes

        # vector graphics (curved paths in general) to images
        if exist_svg:
            excluding_areas = iso_areas
            excluding_areas.extend([block['bbox'] for block in raw['blocks'] if block['type']==1]) # normal images
            images = ImagesExtractor(self.fitz_page) \
                .extract_vector_graphics(excluding_areas, settings['clip_image_res_ratio'])
            raw['blocks'].extend(images)

        # Hyperlink is considered as a Shape
        hyperlinks = self._preprocess_hyperlinks(self.fitz_page)
        raw['shapes'].extend(hyperlinks)

        return paths
    

    @staticmethod
    def _preprocess_hyperlinks(page):
        """Get source hyperlink dicts.

        Args:
            page (fitz.Page): pdf page.

        Returns:
            list: A list of source hyperlink dict.
        """
        hyperlinks = []
        for link in page.getLinks():
            if link['kind']!=2: continue # consider internet address only
            hyperlinks.append({
                'type': RectType.HYPERLINK.value,
                'bbox': tuple(link['from']),
                'uri' : link['uri']
            })

        return hyperlinks


    @staticmethod
    def _create_section(num_col:int, elements:Collection, h_range:tuple, y_ref:float):
        '''Create section based on column count, candidate elements and horizontal boundary.'''
        if not elements: return
        X0, X1 = h_range

        if num_col==1:
            x0, y0, x1, y1 = elements.bbox
            column = Column().update_bbox((X0, y0, X1, y1))
            column.add_elements(elements)
            section = Section(space=0, columns=[column])
            before_space = y0 - y_ref
        else:
            cols = elements.group_by_columns()
            u0, v0, u1, v1 = cols[0].bbox
            m0, n0, m1, n1 = cols[1].bbox
            u = (u1+m0)/2.0

            column_1 = Column().update_bbox((X0, v0, u, v1))
            column_1.add_elements(elements)

            column_2 = Column().update_bbox((u, n0, X1, n1))
            column_2.add_elements(elements)

            section = Section(space=0, columns=[column_1, column_2])
            before_space = v0 - y_ref

        section.before_space = round(before_space, 1)
        return section
                

